"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from typing_extensions import TypeAlias

ID64: TypeAlias = int
ID32: TypeAlias = int

ChatGroupID: TypeAlias = int
ChannelID: TypeAlias = int
