"""
steam.ext._gc
~~~~~~~~~~~~~

A module with common functionality for interacting with the Game Coordinator.

Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE
"""

from .client import *
from .state import *
